@@ -1,5 +1,5 @@
-;;;; $Id: specbot.lisp,v 1.12 2004/11/23 02:54:08 bmastenbrook Exp $
-;;;; $Source: /project/cl-irc/cvsroot/cl-irc/example/specbot.lisp,v $
+;;;; $Id: specbot.lisp 103 2005-08-09 01:26:14Z lisppaste $
+;;;; $Source$
 
 ;;;; specbot.lisp - an example IRC bot for cl-irc
 
@@ -68,6 +68,9 @@
     (elisp-lookup "elisp" "GNU Emacs Lisp Reference Manual")
     (clim-lookup "clim" "Common Lisp Interface Manager II Specification")))
 
+(defvar *spaces-allowed*
+  '(clim-lookup))
+
 (defvar *alists* nil)
 
 (defun add-simple-alist-lookup (file designator prefix description)
@@ -130,7 +133,8 @@
               do
               (aif (strip-address to-lookup :address (second type) :final t)
                    (let ((looked-up (funcall actual-fun it)))
-                     (if (and (<= 0 (count #\space it) 1)
+                     (if (and (<= 0 (count #\space it)
+				  (if (member actual-fun *spaces-allowed*) 1 0)1)
                               (not looked-up))
                          (setf looked-up (format nil "Sorry, I couldn't find anything for ~A."  it)))
                      (and looked-up
@@ -160,10 +164,19 @@
                      (or *load-truename*
                          *default-pathname-defaults*)))))
 
+(defparameter *man-file*
+  (merge-pathnames "man.lisp-expr"
+                   (make-pathname
+                    :directory
+                    (pathname-directory
+                     (or *load-truename*
+                         *default-pathname-defaults*)))))
+
 (defun start-specbot (nick server &rest channels)
   (add-simple-alist-lookup *754-file* 'ieee754 "ieee754" "Section numbers of IEEE 754")
   (add-simple-alist-lookup *ppc-file* 'ppc "ppc" "PowerPC assembly mnemonics")
   (add-simple-alist-lookup *sus-file* 'sus "posix" "Single UNIX Specification")
+  (add-simple-alist-lookup *man-file* 'man "man" "Mac OS X Man Pages")
   (setf *nickname* nick)
   (setf *connection* (connect :nickname *nickname* :server server))
   (mapcar #'(lambda (channel) (join *connection* channel)) channels)
