@@ -1,5 +1,5 @@
-;;;; $Id: protocol.lisp,v 1.14 2005/01/20 23:59:19 ehuelsmann Exp $
-;;;; $Source: /project/cl-irc/cvsroot/cl-irc/protocol.lisp,v $
+;;;; $Id: protocol.lisp 179 2007-04-18 18:57:32Z ehuelsmann $
+;;;; $URL: svn+ssh://ehuelsmann@common-lisp.net/project/cl-irc/svn/trunk/protocol.lisp $
 
 ;;;; See LICENSE for licensing information.
 
@@ -16,6 +16,91 @@
   (:report (lambda (condition stream)
              (format stream "No such reply ~A." (reply-number condition)))))
 
+
+
+;;
+;; Modes
+;;
+
+;; generic abstract mode class
+
+(defclass irc-mode ()
+  ((value
+    :initarg :value
+    :accessor value
+    :initform nil)
+   (value-type
+    :initarg :value-type
+    :accessor value-type
+    :documentation "The framework sets this to `:user' or `:non-user'.
+Essentially, if it's `:user', the value(s) held must be derived from the
+user class.")))
+
+(defgeneric set-mode-value (mode-object value))
+(defgeneric unset-mode-value (mode-object value))
+(defgeneric reset-mode-value (mode-object))
+(defgeneric has-value-p (mode-object value &key key test))
+
+(defmethod reset-mode-value ((mode irc-mode))
+  (setf (value mode) nil))
+
+
+;; mode class for holding boolean values
+
+(defclass boolean-value-mode (irc-mode) ())
+
+(defmethod set-mode-value ((mode boolean-value-mode) value)
+  (declare (ignore value))
+  (setf (value mode) t))
+
+(defmethod unset-mode-value ((mode boolean-value-mode) value)
+  (declare (ignore value))
+  (setf (value mode) nil))
+
+(defmethod has-value-p ((mode boolean-value-mode) value
+                        &key key test)
+  (declare (ignore value key test))
+  (value mode))
+
+;; mode class for holding single values
+
+(defclass single-value-mode (irc-mode) ())
+
+(defmethod set-mode-value ((mode single-value-mode) value)
+  (setf (value mode) value))
+
+(defmethod unset-mode-value ((mode single-value-mode) value)
+  (when (or (null value)
+            (equal value (value mode)))
+    (setf (value mode) nil)))
+
+(defmethod has-value-p ((mode single-value-mode) value
+                        &key (key #'identity) (test #'equal))
+  (funcall test
+           value
+           (funcall key (value mode))))
+
+
+;; mode class for holding lists of values
+
+(defclass list-value-mode (irc-mode) ())
+
+(defmethod set-mode-value ((mode list-value-mode) value)
+  (push value (value mode)))
+
+(defmethod unset-mode-value ((mode list-value-mode) value)
+  (setf (value mode)
+        (remove value (value mode))))
+
+(defmethod has-value-p ((mode list-value-mode) value
+                        &key (key #'identity) (test #'equal))
+  (let ((key-value (funcall key value)))
+    (some #'(lambda (x)
+              (funcall test
+                       key-value
+                       (funcall key x)))
+          (value mode))))
+
 ;;
 ;; Connection
 ;;
@@ -25,14 +110,29 @@
   ((user
     :initarg :user
     :accessor user)
+   (password
+    :initarg :password
+    :accessor password
+    :initform nil)
    (server-name
     :initarg :server-name
     :accessor server-name
     :initform "Unknown server")
-   (server-stream
-    :initarg :server-stream
-    :accessor server-stream
-    :documentation "Stream used to talk to the IRC server.")
+   (server-port
+    :initarg :server-port
+    :accessor server-port
+    :initform *default-irc-server-port*)
+   (socket
+    :initarg :socket
+    :documentation "Slot to store socket (for internal use only).")
+   (network-stream
+    :initarg :network-stream
+    :accessor network-stream
+    :documentation "Stream used to talk binary to the IRC server.")
+   (output-stream
+    :initarg :output-stream
+    :accessor output-stream
+    :documentation "Stream used to send messages to the IRC server")
    (server-capabilities
     :initform *default-isupport-values*
     :accessor server-capabilities
@@ -52,6 +152,25 @@ this stream.")
     :initarg :hooks
     :accessor hooks
     :initform (make-hash-table :test #'equal))
+   (channel-mode-descriptions
+    :initarg :channel-mode-descriptions
+    :accessor channel-mode-descriptions
+    :initform (chanmode-descs-from-isupport *default-isupport-values*)
+    :documentation
+    "Describes the modes an application intends to register with channels.")
+   (nick-prefixes
+    :initarg :nick-prefixes
+    :accessor nick-prefixes
+    :initform (nick-prefixes-from-isupport *default-isupport-values*))
+   (user-mode-destriptions
+    :initarg :user-mode-descriptions
+    :accessor user-mode-descriptions
+    :initform (mapcar #'(lambda (x)
+                          (make-mode-description :char (car x)
+                                                 :symbol (cdr x)))
+                      *char-to-user-modes-map*)
+    :documentation
+    "Describes the modes an application intends to register with channels.")
    (users
     :initarg :users
     :accessor users
@@ -66,12 +185,8 @@ this stream.")
 (defgeneric client-raw-log (connection message))
 (defgeneric connectedp (connection))
 (defgeneric read-message (connection))
-(defgeneric start-process (function name))
-(defgeneric start-background-message-handler (connection))
-(defgeneric read-message-loop (connection))
 (defgeneric read-irc-message (connection))
-(defgeneric send-irc-message (connection command
-                             &optional trailing-argument &rest arguments))
+(defgeneric send-irc-message (connection command &rest arguments))
 (defgeneric get-hooks (connection class))
 (defgeneric add-hook (connection class hook))
 (defgeneric remove-hook (connection class hook))
@@ -81,15 +196,28 @@ this stream.")
 (defgeneric case-map-name (connection))
 (defgeneric re-apply-case-mapping (connection))
 
-(defun make-connection (&key (user nil)
+(defun make-connection (&key (connection-type 'connection)
+                             (user nil)
+                             (password nil)
                              (server-name "")
-                             (server-stream nil)
+                             (server-port nil)
+                             (socket nil)
+                             (network-stream nil)
+                             (outgoing-external-format *default-outgoing-external-format*)
                              (client-stream t)
                              (hooks nil))
-  (let ((connection (make-instance 'connection
+  (let* ((output-stream (flexi-streams:make-flexi-stream
+                         network-stream
+                         :element-type 'character
+                         :external-format (external-format-fixup outgoing-external-format)))
+         (connection (make-instance connection-type
                                    :user user
+                                   :password password
                                    :server-name server-name
-                                   :server-stream server-stream
+                                   :server-port server-port
+                                   :socket socket
+                                   :network-stream network-stream
+                                   :output-stream output-stream
                                    :client-stream client-stream)))
     (dolist (hook hooks)
       (add-hook connection (car hook) (cadr hook)))
@@ -98,9 +226,16 @@ this stream.")
 (defmethod add-default-hooks ((connection connection))
   (dolist (message '(irc-rpl_isupport-message
                      irc-rpl_whoisuser-message
+                     irc-rpl_banlist-message
+                     irc-rpl_endofbanlist-message
+                     irc-rpl_exceptlist-message
+                     irc-rpl_endofexceptlist-message
+                     irc-rpl_invitelist-message
+                     irc-rpl_endofinvitelist-message
                      irc-rpl_list-message
                      irc-rpl_topic-message
                      irc-rpl_namreply-message
+                     irc-rpl_endofnames-message
                      irc-ping-message
                      irc-join-message
                      irc-topic-message
@@ -108,6 +243,8 @@ this stream.")
                      irc-quit-message
                      irc-kick-message
                      irc-nick-message
+                     irc-mode-message
+                     irc-rpl_channelmodeis-message
                      ctcp-time-message
                      ctcp-source-message
                      ctcp-finger-message
@@ -123,34 +260,25 @@ this stream.")
 (defmethod connectedp ((connection connection))
   "Returns t if `connection' is connected to a server and is ready for
 input."
-  (let ((stream (server-stream connection)))
+  (let ((stream (network-stream connection)))
     (and (streamp stream)
          (open-stream-p stream))))
 
-(define-condition invalidate-me (condition)
-  ((stream :initarg :stream
-           :reader invalidate-me-stream)
-   (condition :initarg :condition
-              :reader invalidate-me-condition)))
-
 (defmethod read-message ((connection connection))
-  (let ((read-more-p t))
-    (handler-case
-        (progn
-          (when (and (connectedp connection) read-more-p)
-            (let ((message (read-irc-message connection)))
-              (when *debug-p*
-                (format *debug-stream* "~A" (describe message)))
-              (irc-message-event message)
-              message))) ; needed because of the "loop while" in read-message-loop
-        (stream-error (c) (setf read-more-p nil)
-                    (signal 'invalidate-me :stream
-                            (server-stream connection)
-                            :condition c)))))
+  (when (connectedp connection)
+    (let ((message (read-irc-message connection)))
+      (when *debug-p*
+        (format *debug-stream* "~A" (describe message)))
+      (when message
+        (irc-message-event connection message))
+      message))) ; needed because of the "loop while" in read-message-loop
 
 (defvar *process-count* 0)
 
-(defmethod start-process (function name)
+(defun start-process (function name)
+  "Internal helper for the DEPRECATED function
+START-BACKGROUND-MESSAGE-HANDLER and therefore DEPRECATED itself."
+  (declare (ignorable name))
   #+allegro (mp:process-run-function name function)
   #+cmu (mp:make-process function :name name)
   #+lispworks (mp:process-run-function name nil function)
@@ -158,29 +286,44 @@ input."
   #+openmcl (ccl:process-run-function name function)
   #+armedbear (ext:make-thread function))
 
-(defmethod start-background-message-handler ((connection connection))
+(defun start-background-message-handler (connection)
   "Read messages from the `connection', parse them and dispatch
-irc-message-event on them. Returns background process ID if available."
+irc-message-event on them. Returns background process ID if available.
+
+This function has been DEPRECATED.  The function body is meant as an
+example for library users on handling connection input.  Users
+are strongly encouraged to implement error handling (which is lacking
+from the prototype given here."
+  (warn "START-BACKGROUND-MESSAGE-HANDLER has been deprecated and
+is up for removal in a next release.")
+
+  #+(and sbcl (not sb-thread))
+  (flet ((select-handler (fd)
+            (declare (ignore fd))
+            (if (listen (network-stream connection))
+                (read-message connection)
+              ;; select() returns with no
+              ;; available data if the stream
+              ;; has been closed on the other
+              ;; end (EPIPE)
+              (sb-sys:invalidate-descriptor
+               (sb-sys:fd-stream-fd
+                (network-stream connection))))))
+    (sb-sys:add-fd-handler (sb-sys:fd-stream-fd
+                            (network-stream connection))
+                           :input #'select-handler))
+
+  #-(and sbcl (not sb-thread))
   (flet ((do-loop () (read-message-loop connection)))
     (let ((name (format nil "irc-hander-~D" (incf *process-count*))))
-      #+(or allegro cmu lispworks sb-thread openmcl armedbear)
-      (start-process #'do-loop name)
-      #+(and sbcl (not sb-thread))
-      (sb-sys:add-fd-handler (sb-sys:fd-stream-fd
-			      (server-stream connection))
-			     :input (lambda (fd)
-				      (declare (ignore fd))
-                                      (handler-case
-                                          (read-message connection)
-                                        (invalidate-me (c)
-                                          (sb-sys:invalidate-descriptor
-                                           (sb-sys:fd-stream-fd
-                                            (invalidate-me-stream c)))
-                                          (format t "Socket closed: ~A~%"
-                                                  (invalidate-me-condition c)))))))))
+      (start-process #'do-loop name))))
 
 (defun stop-background-message-handler (process)
-  "Stops a background message handler process returned by the start function."
+  "Stops a background message handler process returned by the start function.
+
+Just as its cousin START-BACKGROUND-MESSAGE-HANDLER,
+this function is DEPRECATED."
+  (declare (ignorable process))
     #+cmu (mp:destroy-process process)
     #+allegro (mp:process-kill process)
     #+sb-thread (sb-thread:destroy-thread process)
@@ -188,25 +331,52 @@ irc-message-event on them. Returns background process ID if available."
     #+openmcl (ccl:process-kill process)
     #+armedbear (ext:destroy-thread process))
 
-(defmethod read-message-loop ((connection connection))
+(defun read-message-loop (connection)
   (loop while (read-message connection)))
 
+(defun try-decode-line (line external-formats)
+  (loop for external-format in external-formats
+        for decoded = nil
+        for error = nil
+        do (multiple-value-setq (decoded error)
+             (handler-case
+              (flexi-streams:with-input-from-sequence (in line)
+                (let ((flexi (flexi-streams:make-flexi-stream in
+;;                                                              :element-type 'character
+                                                              :external-format
+                                                              (external-format-fixup external-format))))
+                  (read-line flexi nil nil)))
+              (flexi-streams:flexi-stream-encoding-error ()
+                  nil)))
+        if decoded
+        do (return decoded)))
+
 (defmethod read-irc-message ((connection connection))
   "Read and parse an IRC-message from the `connection'."
-  (let ((message (create-irc-message
-                  (read-line (server-stream connection) t))))
-    (setf (connection message) connection)
-    message))
+  (handler-case
+   (multiple-value-bind
+       (buf buf-len)
+       ;; Note: we cannot use read-line here (or any other
+       ;; character based functions), since they may cause conversion
+       (read-sequence-until (network-stream connection)
+                            (make-array 1024
+                                        :element-type '(unsigned-byte 8)
+                                        :fill-pointer t)
+                            '(13 10))
+     (setf (fill-pointer buf) buf-len)
+     (let* ((message (create-irc-message (try-decode-line buf *default-incoming-external-formats*))))
+       (setf (connection message) connection)
+       message))
+    (end-of-file ())))
+       ;; satisfy read-message-loop assumption of nil when no more messages
 
 (defmethod send-irc-message ((connection connection) command
-                             &optional trailing-argument &rest arguments)
+                             &rest arguments)
   "Turn the arguments into a valid IRC message and send it to the
 server, via the `connection'."
-  (let ((raw-message (make-irc-message command
-                                       :arguments arguments
-                                       :trailing-argument trailing-argument)))
-    (write-sequence raw-message (server-stream connection))
-    (force-output (server-stream connection))
+  (let ((raw-message (apply #'make-irc-message command arguments)))
+    (write-sequence raw-message (output-stream connection))
+    (force-output (output-stream connection))
     raw-message))
 
 (defmethod get-hooks ((connection connection) (class symbol))
@@ -256,7 +426,7 @@ server, via the `connection'."
       (maphash #'(lambda (norm-name channel)
                    (declare (ignore norm-name))
                    (setf (gethash
-                          (setf (normalized-channel-name channel)
+                          (setf (normalized-name channel)
                                 (normalize-channel-name connection
                                                         (name channel)))
                           new-channels) channel)
@@ -275,18 +445,13 @@ server, via the `connection'."
     :accessor user
     :documentation "The user at the other end of this connection.  The
 user at this end can be reached via your normal connection object.")
-   (stream
-    :initarg :stream
-    :accessor dcc-stream)
+   (network-stream
+    :initarg :network-stream
+    :accessor network-stream)
    (output-stream
     :initarg :output-stream
     :accessor output-stream
-    :initform t)
-   (socket
-    :initarg :socket
-    :accessor socket
-    :documentation "The actual socket object for the connection
-between the two users.")))
+    :initform t)))
 
 (defmethod print-object ((object dcc-connection) stream)
   "Print the object for the Lisp reader."
@@ -301,50 +466,42 @@ between the two users.")))
                                  (remote-address nil)
                                  (remote-port nil)
                                  (output-stream t))
-  #+sbcl
-  (let ((socket (sb-bsd-sockets:make-inet-socket :stream :tcp)))
-    (sb-bsd-sockets:socket-connect socket remote-address remote-port)
-    (make-instance 'dcc-connection
-                   :user user
-                   :stream (sb-bsd-sockets:socket-make-stream socket :input t :output t :buffering :none)
-                   :socket socket
-                   :output-stream t))
-  #+openmcl
-  (let ((socket-stream (ccl:make-socket :remote-host remote-address
-                                        :remote-port remote-port)))
-    (make-instance 'dcc-connection
-                   :user user
-                   :stream socket-stream
-                   :output-stream output-stream))
-  #-(or openmcl sbcl)
-  (warn "make-dcc-connection not supported for this implementation."))
+  (make-instance 'dcc-connection
+                 :user user
+                 :network-stream (usocket:socket-connect remote-address
+                                                         remote-port)
+                 :output-stream output-stream))
 
 (defgeneric dcc-close (connection))
 (defgeneric send-dcc-message (connection message))
 
 (defmethod read-message ((connection dcc-connection))
-  (let ((message (read-line (dcc-stream connection))))
-    (format (output-stream connection) "~A~%" message)
-    (force-output (output-stream connection))
-    message))
-
-(defmethod read-message-loop ((connection dcc-connection))
-  (loop while (read-message connection)))
+  (when (connectedp connection)
+    (let ((message (read-line (network-stream connection))))
+      (format (output-stream connection) "~A~%" message)
+      (force-output (output-stream connection))
+      (when *debug-p*
+        (format *debug-stream* "~A" (describe message)))
+      ;; (dcc-message-event message)
+      message))) ; needed because of the "loop while" in read-message-loop
 
 (defmethod send-dcc-message ((connection dcc-connection) message)
-  (format (dcc-stream connection) "~A~%" message))
+  (format (network-stream connection) "~A~%" message)
+  (force-output (network-stream connection)))
 
 ;; argh.  I want to name this quit but that gives me issues with
 ;; generic functions.  need to resolve.
 (defmethod dcc-close ((connection dcc-connection))
-  (close (dcc-stream connection))
+  #+(and sbcl (not sb-thread))
+  (sb-sys:invalidate-descriptor
+   (sb-sys:fd-stream-fd (network-stream connection)))
+  (close (network-stream connection))
   (setf (user connection) nil)
   (setf *dcc-connections* (remove connection *dcc-connections*))
-  #+sbcl (sb-bsd-sockets:socket-close (socket connection))
   )
 
 (defmethod connectedp ((connection dcc-connection))
-  (let ((stream (dcc-stream connection)))
+  (let ((stream (network-stream connection)))
     (and (streamp stream)
          (open-stream-p stream))))
 
@@ -365,7 +522,12 @@ between the two users.")))
    (modes
     :initarg :modes
     :accessor modes
-    :initform nil)
+    :initform '())
+   (visibility
+    :initarg :visibility
+    :accessor visibility
+    :initform nil
+    :type (member nil :public :private :secret :unknown))
    (users
     :initarg :users
     :accessor users
@@ -414,6 +576,47 @@ name."
 (defgeneric remove-channel (connection channel))
 (defgeneric remove-users (channel))
 
+(defgeneric mode-name-from-char (connection target mode-char)
+  (:documentation "Map the mode character used in the MODE message to a
+symbol used internally to describe the mode given a `target'."))
+
+(defgeneric mode-description (connection target mode-name)
+  (:documentation "Retrieve a `mode-description' structure for the given
+`mode-name' keyword."))
+
+(defgeneric get-mode (target mode)
+  (:documentation "Get the value associated with `mode' for `target'
+or `nil' if no mode available."))
+
+(defgeneric set-mode (target mode &optional parameter)
+  (:documentation "Set the mode designated by the `mode' keyword to a
+value passed in `parameter' or T if `parameter' is absent."))
+
+(defgeneric unset-mode (target mode &optional parameter)
+  (:documentation
+"Sets value of the mode designated by the `mode' keyword to nil.
+If the mode holds a list of values `parameter' is used to indicate which
+element to remove."))
+
+(defgeneric add-mode (target mode-name mode)
+  (:documentation "Add the mode-holding object `mode-value' to `target'
+under the access key `mode-name'.
+
+If mode-value is a subtype of irc-mode, it is added as-is.
+Otherwise, a mode-object will be generated from the "))
+(defgeneric remove-mode (target mode-name)
+  (:documentation "Remove the mode-holding object in the `mode-name' key
+from `target'."))
+
+(defgeneric has-mode-p (target mode)
+  (:documentation "Return a generalised boolean indicating if `target' has
+a mode `mode' associated with it."))
+
+(defgeneric has-mode-value-p (target mode value &key key test)
+  (:documentation "Return a generalised boolean indicating if `target' has
+a mode `mode' associated with the value `value' for given a `key' transform
+and `test' test."))
+
 (defmethod find-channel ((connection connection) (channel string))
   "Return channel as designated by `channel'.  If no such channel can
 be found, return nil."
@@ -434,7 +637,68 @@ be found, return nil."
 
 (defmethod remove-users ((channel channel))
   "Remove all users on `channel'."
-  (clrhash (users channel)))
+  (clrhash (users channel))
+  (do-property-list (prop val (modes channel))
+     (when (and val (eq (value-type val) :user))
+       (remf (modes channel) prop))))
+
+(defmethod mode-name-from-char ((connection connection)
+                                (target channel) mode-char)
+  (declare (ignore target))
+  (let ((mode-desc (find mode-char (channel-mode-descriptions connection)
+                         :key #'mode-desc-char)))
+    (when mode-desc
+      (mode-desc-symbol (the mode-description mode-desc)))))
+
+(defmethod mode-description ((connection connection)
+                             (target channel) mode-name)
+  (declare (ignore target))
+  (find mode-name (channel-mode-descriptions connection)
+        :key #'mode-desc-symbol))
+
+(defgeneric make-mode (connection target mode-id))
+
+(defmethod make-mode (connection target (mode character))
+  (let ((mode-name (mode-name-from-char connection target mode)))
+    (make-mode connection target mode-name)))
+
+(defmethod make-mode (connection target (mode symbol))
+  (let ((mode-desc (mode-description connection target mode)))
+    (make-instance (mode-desc-class mode-desc)
+                   :value-type (if (mode-desc-nick-param-p mode-desc)
+                                   :user :non-user))))
+
+(defmethod add-mode (target mode-name mode)
+  (setf (getf (modes target) mode-name) mode))
+
+(defmethod remove-mode (target mode-name)
+  (remf (modes target) mode-name))
+
+(defmethod get-mode (target mode)
+  (let ((mode-object (has-mode-p target mode)))
+    (when mode-object
+      (value mode-object))))
+
+(defmethod set-mode (target mode &optional parameter)
+  (set-mode-value (getf (modes target) mode) parameter))
+
+(defmethod unset-mode (target mode &optional parameter)
+  (let ((mode (getf (modes target) mode)))
+    (when mode
+      (unset-mode-value mode parameter))))
+
+(defmethod has-mode-p (target mode)
+  (multiple-value-bind
+      (indicator value tail)
+      (get-properties (modes target) (list mode))
+    (when (or indicator value tail)
+      value)))
+
+(defmethod has-mode-value-p (target mode value
+                                    &key (key #'identity) (test #'equal))
+  (let ((mode (getf (modes target) mode)))
+    (when mode
+      (has-value-p mode value :key key :test test))))
 
 ;;
 ;; User
@@ -461,6 +725,10 @@ be found, return nil."
     :initarg :realname
     :accessor realname
     :initform "")
+   (modes
+    :initarg :modes
+    :accessor modes
+    :initform '())
    (channels
     :initarg :channels
     :accessor channels
@@ -493,7 +761,7 @@ be found, return nil."
              (second (assoc "PREFIX"
                             (server-capabilities connection)
                             :test #'string=))))
-      (subseq nickname 1)
+      (substring nickname 1)
       nickname))
 
 (defun normalize-nickname (connection string)
@@ -534,15 +802,10 @@ known."
 (defmethod remove-user ((channel channel) (user user))
   "Remove `user' from `channel' and `channel' from `user'."
   (remhash (normalized-nickname user) (users channel))
-  (setf (channels user) (remove channel (channels user))))
-
-(defmethod remove-channel ((channel channel) (user user))
-  "Remove `channel' from `user'."
-  (warn
-   (concatenate 'string
-                "use of depricated API (remove-channel channel user): "
-                "(remove-channel user channel) is now preferred"))
-  (remove-channel user channel))
+  (setf (channels user) (remove channel (channels user)))
+  (do-property-list (prop val (modes channel))
+     (when (and val (eq (value-type val) :user))
+       (unset-mode channel prop user))))
 
 (defmethod remove-channel ((user user) (channel channel))
   "Remove `channel' from `user'."
@@ -559,27 +822,50 @@ may be already be on."
     (remove-user channel user))
   (remove-user connection user))
 
+(defmethod mode-name-from-char ((connection connection)
+                                (target user) mode-char)
+  (declare (ignore target))
+  (let ((mode-desc (find mode-char (user-mode-descriptions connection)
+                         :key #'mode-desc-char)))
+    (when mode-desc
+      (mode-desc-symbol (the mode-description mode-desc)))))
+
+(defmethod mode-description ((connection connection)
+                             (target user) mode-name)
+  (declare (ignore target))
+  (find mode-name (user-mode-descriptions connection)
+        :key #'mode-desc-symbol))
+
 (defmethod find-or-make-user ((connection connection) nickname &key (username "")
                               (hostname "") (realname ""))
-  (or (find-user connection nickname)
-      (make-user connection
-                 :nickname nickname
-                 :username username
-                 :hostname hostname
-                 :realname realname)))
+  (let ((user (find-user connection nickname)))
+    (unless user
+      (setf user
+            (make-user connection
+                       :nickname nickname
+                       :username username
+                       :hostname hostname
+                       :realname realname)))
+    (labels ((update-slot-if-known (slotname value)
+               (when (string= (slot-value user slotname) "")
+                 (setf (slot-value user slotname) value))))
+      (update-slot-if-known 'username username)
+      (update-slot-if-known 'hostname hostname)
+      (update-slot-if-known 'realname realname))
+    user))
 
 (defmethod change-nickname ((connection connection) (user user) new-nickname)
-  (let ((new-user user)
-        (channels (channels user)))
+  (let ((channels (channels user)))
     (remove-user connection user)
-    (setf (nickname new-user) new-nickname)
-    (setf (normalized-nickname new-user)
+    (dolist (channel channels)
+      (remove-user channel user))
+    (setf (nickname user) new-nickname)
+    (setf (normalized-nickname user)
           (normalize-nickname connection new-nickname))
     (dolist (channel channels)
-      (remove-user channel user)
-      (add-user channel new-user))
+      (add-user channel user))
     (add-user connection user)
-    new-user))
+    user))
 
 ;; IRC Message
 ;;
@@ -604,10 +890,6 @@ may be already be on."
     :accessor arguments
     :initarg :arguments
     :type list)
-   (trailing-argument
-    :accessor trailing-argument
-    :initarg :trailing-argument
-    :type string)
    (connection
     :accessor connection
     :initarg :connection)
@@ -624,7 +906,15 @@ may be already be on."
   (print-unreadable-object (object stream :type t :identity t)
     (format stream "~A ~A" (source object) (command object))))
 
+;;Compat code; remove after 2006-08-01
+
+(defgeneric trailing-argument (message))
+(defmethod trailing-argument ((message irc-message))
+  (warn "Use of deprecated function irc:trailing-argument")
+  (car (last (arguments message))))
+
 (defgeneric self-message-p (message))
+(defgeneric user-eq-me-p (connection user))
 (defgeneric find-irc-message-class (type))
 (defgeneric client-log (connection message &optional prefix))
 (defgeneric apply-to-hooks (message))
@@ -634,6 +924,13 @@ may be already be on."
   (string-equal (source message)
                 (nickname (user (connection message)))))
 
+(defmethod user-eq-me-p (connection (user user))
+  (eq user (user connection)))
+
+(defmethod user-eq-me-p (connection (user string))
+  (let ((user (find-user connection user)))
+    (user-eq-me-p connection user)))
+
 (defclass irc-error-reply (irc-message) ())
 
 (eval-when (:compile-toplevel :load-toplevel :execute)
@@ -641,19 +938,19 @@ may be already be on."
     "Intern based on symbol-name to support case-sensitive mlisp"
     (intern
      (concatenate 'string
-		  (symbol-name prefix)
-		  "-"
-		  (symbol-name name)
-		  "-"
-		  (symbol-name '#:message))))
+                  (symbol-name prefix)
+                  "-"
+                  (symbol-name name)
+                  "-"
+                  (symbol-name '#:message))))
 
   (defun define-irc-message (command)
     (let ((name (intern-message-symbol :irc command)))
       `(progn
-	(defmethod find-irc-message-class ((type (eql ,command)))
-	  (find-class ',name))
-	(export ',name)
-	(defclass ,name (irc-message) ())))))
+        (defmethod find-irc-message-class ((type (eql ,command)))
+          (find-class ',name))
+        (export ',name)
+        (defclass ,name (irc-message) ())))))
 
 (defmacro create-irc-message-classes (class-list)
   `(progn ,@(mapcar #'define-irc-message class-list)))
@@ -661,7 +958,7 @@ may be already be on."
 ;; should perhaps wrap this in an eval-when?
 (create-irc-message-classes #.(remove-duplicates (mapcar #'second *reply-names*)))
 (create-irc-message-classes (:privmsg :notice :kick :topic :error :mode :ping
-			     :nick :join :part :quit :kill :pong :invite))
+                             :nick :join :part :quit :kill :pong :invite))
 
 (defmethod find-irc-message-class (type)
   (declare (ignore type))
@@ -674,8 +971,8 @@ may be already be on."
             (received-time message)
             (command message)
             (source message)
-            (arguments message)
-            (trailing-argument message))
+            (butlast (arguments message))
+            (car (last (arguments message))))
     (force-output stream)))
 
 (defmethod apply-to-hooks ((message irc-message))
@@ -695,15 +992,17 @@ may be already be on."
 (defclass standard-ctcp-message (ctcp-mixin irc-message) ())
 
 (defgeneric find-ctcp-message-class (type))
+(defgeneric ctcp-request-p (message))
+(defgeneric ctcp-reply-p (message))
 
 (eval-when (:compile-toplevel :load-toplevel :execute)
   (defun define-ctcp-message (ctcp-command)
     (let ((name (intern-message-symbol :ctcp ctcp-command)))
       `(progn
-	(defmethod find-ctcp-message-class ((type (eql ,ctcp-command)))
-	  (find-class ',name))
-	(export ',name)
-	(defclass ,name (ctcp-mixin irc-message) ())))))
+        (defmethod find-ctcp-message-class ((type (eql ,ctcp-command)))
+          (find-class ',name))
+        (export ',name)
+        (defclass ,name (ctcp-mixin irc-message) ())))))
 
 (defmacro create-ctcp-message-classes (class-list)
   `(progn ,@(mapcar #'define-ctcp-message class-list)))
@@ -717,6 +1016,12 @@ may be already be on."
   (declare (ignore type))
   (find-class 'standard-ctcp-message))
 
+(defmethod ctcp-request-p ((message ctcp-mixin))
+  (string= (command message) :privmsg))
+
+(defmethod ctcp-reply-p ((message ctcp-mixin))
+  (string= (command message) :notice))
+
 (defmethod client-log ((connection connection) (message ctcp-mixin) &optional (prefix ""))
   (let ((stream (client-stream connection)))
     (format stream "~A~A: ~A (~A): ~A~{ ~A~} \"~A\"~%"
@@ -725,7 +1030,7 @@ may be already be on."
             (command message)
             (ctcp-command message)
             (source message)
-            (arguments message)
-            (trailing-argument message))
+            (butlast (arguments message))
+            (car (last (arguments message))))
     (force-output stream)))
 
