@@ -1,5 +1,5 @@
-;;;; $Id: variable.lisp,v 1.3 2005/01/20 23:59:19 ehuelsmann Exp $
-;;;; $Source: /project/cl-irc/cvsroot/cl-irc/variable.lisp,v $
+;;;; $Id: variable.lisp 160M 2007-04-19 21:22:46Z (local) $
+;;;; $URL: svn+ssh://ehuelsmann@common-lisp.net/project/cl-irc/svn/trunk/variable.lisp $
 
 ;;;; See the LICENSE file for licensing information.
 
@@ -10,7 +10,7 @@
 
 (defconstant +soh+ #.(code-char 1))
 
-(defparameter *version* "0.1.0")
+(defparameter *version* "0.8.1")
 (defparameter *ctcp-version*
   (format nil "CL IRC library, cl-irc:~A:~A ~A"
           *version* (machine-type) (machine-version)))
@@ -26,18 +26,85 @@
 (defvar *default-quit-message*
   "Common Lisp IRC library - http://common-lisp.net/project/cl-irc")
 
+(defparameter *default-isupport-CHANMODES*
+  "beI,kO,l,aimnpqsrt")
+(defparameter *default-isupport-PREFIX*
+  "(ov)@+")
+
 (defparameter *default-isupport-values*
-  '(("CASEMAPPING" "rfc1459")
+  `(("CASEMAPPING" "rfc1459")
+    ("CHANMODES" ,*default-isupport-CHANMODES*)
     ("CHANNELLEN" "200")
     ("CHANTYPES" "#&")
     ("MODES" "3")
     ("NICKLEN" "9")
-    ("PREFIX" "(ov)@+")
+    ("PREFIX" ,*default-isupport-PREFIX*)
     ("TARGMAX")))
 
+(defparameter *default-outgoing-external-format* '(:utf-8)
+  "The external-format we use to encode outgoing messages. This
+  should be an external format spec that flexi-streams accepts.
+
+  :eol-style will always be overridden to be :crlf as required
+  by the IRC protocol.")
+
+(defparameter *default-incoming-external-formats* '((:utf-8 :eol-style :crlf)
+                                                    (:latin1 :eol-style :crlf))
+  "The external-formats we use to decode incoming messages. This should
+  be a list of external format specs that flexi-streams accepts.
+
+  The external formats are tried in order, until one decodes the
+  message without encoding errors. Note that the last external
+  format should be a single-byte one with most or even all valid
+  codepoints (such as latin-1).
+
+  :eol-style will always be overridden to be :crlf as required by the
+  IRC protocol.")
 
 (defvar *dcc-connections* nil)
 
+(defstruct (mode-description (:conc-name "MODE-DESC-"))
+  (char nil)
+  (symbol nil)
+  (param-on-set-p nil)
+  (param-on-unset-p nil)
+  (nick-param-p nil)
+  (class 'single-value-mode))
+
+
+(defparameter *default-char-to-channel-modes-map*
+  '(
+    ;; these modes don't take parameters
+    (#\a . :anonymous)
+    (#\i . :invite-only)
+    (#\m . :moderated)
+    (#\n . :no-external)
+    (#\q . :quiet)
+    (#\s . :secret)
+    (#\r . :reop)
+    (#\t . :op-only-topic)
+
+    ;; these modes take a user parameter
+    (#\O . :channel-creator)
+    (#\o . :channel-operator)
+    (#\v . :voice)
+
+    ;; these modes take a parameter other than a user
+    (#\l . :limit)
+    (#\k . :key)
+    (#\b . :ban)
+    (#\e . :except)
+    (#\I . :invite)))
+
+(defparameter *char-to-user-modes-map*
+  '((#\a . :away)
+    (#\i . :invisible)
+    (#\w . :receive-wallops)
+    (#\s . :server-notices)
+    (#\r . :restricted-connection)
+    (#\o . :remote-operator)
+    (#\O . :local-operator)))
+
 (defparameter *reply-names*
   '((1 :rpl_welcome)
     (2 :rpl_yourhost)
@@ -46,6 +113,12 @@
     (5 :rpl_isupport) ;; The RFC was wrong to define RPL_BOUNCE here,
      ;; see http://www.irc.org/tech_docs/draft-brocklesby-irc-isupport-03.txt
     (10 :rpl_bounce)
+    (15 :rpl_map) ; From ircd 2.11 source
+    (17 :rpl_mapend) ; From ircd 2.11 source
+    (18 :rpl_mapstart) ; From ircd 2.11 source
+    (20 :rpl_hello) ; From ircd 2.11 source
+    (42 :rpl_yourid) ; From ircd 2.11 source
+    (43 :rpl_savenick) ; From ircd 2.11 source
     (200 :rpl_tracelink)
     (201 :rpl_traceconnecting)
     (202 :rpl_tracehandshake)
